# -*- coding:utf-8 -*-
# CREATED BY: bohuai jiang 
# CREATED ON: 2020/8/7 3:50 PM
# LAST MODIFIED ON:
# AIM:
from abc import ABC
import re

from sentence_spliter.automata.abc import Criteria
from sentence_spliter.automata.sequence import StrSequence
from sentence_spliter.automata.symbols import SYMBOLS


class IsEndState(Criteria):
    def __init__(self):
        super(IsEndState, self).__init__('IsEndState')

    def accept(self, state: StrSequence) -> bool:
        return state.reach_right_end()


class IsEndSymbol(Criteria):
    def __init__(self):
        super(IsEndSymbol, self).__init__('IsEndSymbolZH')
        self.empty = re.compile('\s')

    def is_zh(self, char: str):
        if '\u4e00' <= char <= '\u9fa5':
            return True

    def accept(self, state: StrSequence) -> bool:
        if SYMBOLS['end_symbols'].match(state.current_value) and \
                not (SYMBOLS['en_dot'].match(state.next_value) or SYMBOLS['end_symbols'].match(state.next_value)):
            return True
        if SYMBOLS['en_dot'].match(state.current_value):
            if self.is_zh(state.next_value) or self.empty.match(state.next_value):
                return True
        return False


class IsComma(Criteria):
    def __init__(self):
        super(IsComma, self).__init__('IsComma')

    def accept(self, state: StrSequence) -> bool:
        if SYMBOLS['comma'].match(state.current_value):
            return True
        return False


class IsBracketClose(Criteria):
    def __init__(self):
        super(IsBracketClose, self).__init__('IsBracketClose')

    def accept(self, state: StrSequence) -> bool:
        # -- avoid endless loop -- #
        if state.reach_right_end():
            return True
        if state.bracket_left <= state.bracket_right:
            state.reset_bracket()
            return True
        else:
            return False


class IsQuoteClose(Criteria):
    def __init__(self):
        super(IsQuoteClose, self).__init__('IsQuoteClose')

    def accept(self, state: StrSequence) -> bool:
        # -- avoid endless loop -- #
        if state.reach_right_end():
            return True

        if state.quota_en > 0:
            if (state.quota_en + state.quota_left) % 2 == 0:
                state.reset_quota()
                return True
            else:
                return False
        elif state.quota_left > 0:
            return False
        else:
            return True


class IsBookClose(Criteria):
    def __init__(self):
        super(IsBookClose, self).__init__('IsBookClose')

    def accept(self, state: StrSequence) -> bool:
        # -- avoid endless loop -- #
        if state.reach_right_end():
            return True
        if state.book_left <= state.book_right:
            # state.reset_bookmark()
            return True
        else:
            return False


class IsLongSentence(Criteria):
    def __init__(self, max_len: int = 128):
        super(IsLongSentence, self).__init__('IsLongSentence')
        self.max_len = max_len

    def accept(self, state: StrSequence) -> bool:
        if state.candidate_len > self.max_len:
            return True
        return False


class IsShortSentence(Criteria):
    def __init__(self, min_len: int = 17):
        super(IsShortSentence, self).__init__('IsShortSentence')
        self.min_len = min_len

    def accept(self, state: StrSequence) -> bool:
        if state.candidate_len < self.min_len:
            return True
        return False


class IsSpace(Criteria):
    def __init__(self):
        super(IsSpace, self).__init__('IsSpace')
        self.pattern = re.compile('\s')

    def accept(self, state: StrSequence) -> bool:
        if self.pattern.match(state.current_value):
            return True
        else:
            return False


# =========================== #
#      Special Condition      #
# =========================== #

class SpecialCondition(Criteria, ABC):
    pass


class IsRightQuota(SpecialCondition):
    def __init__(self, index=0):
        super(IsRightQuota, self).__init__('IsRightQuota')
        self.index = index

    def accept(self, state: StrSequence) -> bool:
        if self.index >= 0:
            index = min(state.v_pointer + self.index, state.length - 1)
        else:
            index = max(0, state.v_pointer + self.index)
        if SYMBOLS["quotation_right"].match(state[index]):
            return True
        return False


class IsLeftQuota(SpecialCondition):
    def __init__(self, index=0):
        super(IsLeftQuota, self).__init__('IsLeftQuota')
        self.index = index

    def accept(self, state: StrSequence) -> bool:
        if self.index >= 0:
            index = min(state.v_pointer + self.index, state.length - 1)
        else:
            index = max(0, state.v_pointer + self.index)
        if SYMBOLS["quotation_left"].match(state[index]):
            return True
        return False


class IsLeftQuotaGreaterThan(SpecialCondition):
    '''
    左引号 重复次数大于一定值
    '''

    def __init__(self, theta: int = 1):
        super(IsLeftQuotaGreaterThan, self).__init__('IsLeftQuotaGreaterThan')
        self.theta = theta

    def accept(self, state: StrSequence) -> bool:
        if state.quota_left > self.theta:
            state.reset_quota()
            state.v_pointer -= 1
            return True


class IsRQuotaStickWithLQuota(SpecialCondition):
    '''
    右引号紧跟着左引号
    '''

    def __init__(self):
        super(IsRQuotaStickWithLQuota, self).__init__('IsRQuotaStickWithLQuota')

    def accept(self, state: StrSequence) -> bool:
        if SYMBOLS['quotation_right'].match(state.current_value) and \
                SYMBOLS['quotation_left'].match(state.next_value):
            return True
        return False


class WithSays(SpecialCondition):
    def __init__(self):
        super(WithSays, self).__init__('WithSays')
        self.pattern = re.compile('([他她](说|说道|笑道|道))+')

    def accept(self, state: StrSequence) -> bool:
        sentence = state[state.v_pointer + 1:15]
        if self.pattern.match(sentence):
            return True
        return False


class SpecialEnds(SpecialCondition):
    def __init__(self):
        super(SpecialEnds, self).__init__('SpecialEnds')
        self.pattern_pre = re.compile('([~～])')
        self.pattern_cur = re.compile('\s')

    def accept(self, state: StrSequence) -> bool:
        if self.pattern_pre.match(state.pre_value) and self.pattern_cur.match(state.current_value):
            return True
        else:
            return False



# -*- coding:utf-8 -*-
# CREATED BY: bohuai jiang 
# CREATED ON: 2020/8/12 10:30 AM
# LAST MODIFIED ON:
# AIM: 状态机器

from automata.state_machine import StateMachine

from automata.sequence import StrSequence
from sentence_spliter.logic_graph import long_short_cuter, simple_cuter, special_cuter
from sentence_spliter.spliter import cut_to_sentences
machine = StateMachine(special_cuter())
#machine = StateMachine(special_graph)
# machine = StateMachine(logic_simple)
if __name__ == '__main__':
    novel = '/Users/Shared/小说/女频小说25本-Estella/女频小说25本-Estella/你原来是这样的叶先生.txt'
    with open(novel, 'r') as f:
        data = f.read()

    sentence_org = '在很久很久以前......... 有座山。山里有座庙啊!!!庙里竟然有个老和尚！？。。。。www.baidu.com.1.2.3'
    #'“我和你讨论的不是一个东西，死亡率与死亡比例是不同的” “你知道么？CNN你们总是制造假新闻。。。”'
    sentence = StrSequence(sentence_org)


    # sentence = StrSequence('《掏出绝命岛!》真好看。')
    machine.run(sentence, verbose=True)
    print('\n'.join(sentence.sentence_list()))



# -*- coding:utf-8 -*-
# CREATED BY: bohuai jiang 
# CREATED ON: 2020/8/16 5:34 PM
# LAST MODIFIED ON:
# AIM:
import sys
from typing import Union, List

from sentence_spliter.automata.symbols import SYMBOLS


def print_percent(current: int, max: int, header: str = ''):
    percent = float(current) / max * 100
    sys.stdout.write("\r{0}{1:.3g}%".format(header, percent))


class StrSequence:
    def __init__(self, str_block: Union[str, List[str]], verbose=None):
        self.value = str_block
        # -- pointers -- #
        self.v_pointer = -1
        self.sentence_start = 0
        # -- results -- #
        self.sentence_list_idx = []
        self.length = len(self.value)
        self.verbose = verbose
        if verbose is None and len(self.value) > 50000:
            self.verbose = True

        # -- other pointers -- #
        self.bracket_left = 0
        self.bracket_right = 0
        self.quota_left = 0
        self.quota_en = 0
        self.book_left = 0
        self.book_right = 0

    # -- magic method -- #
    def __str__(self):
        return self.value[self.sentence_start:self.v_pointer]

    def __getitem__(self, item: int):
        return self.value[item]

    @property
    def pre_value(self):
        return self[max(0, self.v_pointer - 1)]

    @property
    def next_value(self):
        return self[min(self.length - 1, self.v_pointer + 1)]

    @property
    def candidate_len(self):
        return self.v_pointer - self.sentence_start + 1

    @property
    def current_value(self):
        if self.value:
            try:
                return self.value[self.v_pointer]
            except:
                return self.value[-1]
        else:
            return self.value

    def sentence_list(self):
        out = []
        for v in self.__sentence_list:
            if v[0] < v[1]:
                if isinstance(self.value, str):
                    out.append(self.value[v[0]:v[1]])
                else:
                    out.append(' '.join(self.value[v[0]:v[1]]))
        return out

    def at_end_pos(self):
        return self.v_pointer >= self.length

    def add_to_candidate(self):
        self.v_pointer = min(self.v_pointer + 1, self.length)
        # -- update pointer -- #
        self.update_barcket()
        self.update_quota()
        self.update_bookmark()
        if self.verbose:
            print_percent(self.v_pointer, self.length, 'process cutting ')

    def add_to_sentence_list(self):
        self.__sentence_list.append((self.sentence_start, min(self.length, self.v_pointer + 1)))
        self.sentence_start = self.v_pointer + 1

    def reach_right_end(self):
        if self.v_pointer >= self.length - 1:
            self.v_pointer = self.length - 1
            return True
        else:
            return False

    # --- update pointer --- #
    def update_barcket(self):
        key = self.current_value
        if SYMBOLS['bracket_left'].match(key):
            self.bracket_left += 1
        if SYMBOLS['bracket_right'].match(key):
            self.bracket_right += 1

    def reset_bracket(self):
        self.bracket_left = 0
        self.bracket_right = 0

    def update_quota(self):
        word = self.current_value
        if SYMBOLS['quotation_en'].match(word):
            self.quota_en += 1
        if SYMBOLS['quotation_left'].match(word):
            self.quota_left += 1
        if SYMBOLS['quotation_right'].match(word):
            self.quota_en = 0
            self.quota_left = 0

    def reset_quota(self):
        self.quota_en = 0
        self.quota_left = 0

    def update_bookmark(self):
        word = self.current_value
        if SYMBOLS['book_left'].match(word):
            self.book_left += 1
        if SYMBOLS['book_right'].match(word):
            self.book_right += 1

    def reset_bookmark(self):
        self.book_left = 0
        self.book_right = 0



# -*- coding:utf-8 -*-
# CREATED BY: bohuai jiang 
# CREATED ON: 2020/8/12 11:17 AM
# LAST MODIFIED ON:
# AIM: 构建状态机
from typing import Set, Callable

from sentence_spliter.automata.abc import Criteria, Operation
from sentence_spliter.automata.sequence import StrSequence


class StateMachine:
    def __init__(self, graph: dict):
        self.graph = graph
        if not self.__is_valid_graph():
            raise Exception('invalid graph, please check if all nodes can connect to end_state')

        self.init_state = self.get_init_state()

    def get_init_state(self):
        for key in self.graph:
            if key.name == "Indolent":
                return key
        else:
            raise Exception('invalid graph, init_State not found')

    def __is_valid_graph(self) -> bool:
        for node in self.__get_all_nodes():
            try:
                for child in self.graph[node]:
                    if child['node'].name == 'EndState':
                        break
                else:
                    return False
            except:
                return False
        return True

    def __get_all_nodes(self) -> Set[Operation]:
        out = set()
        for node in self.graph:
            if node.name != 'EndState':
                out.add(node)
            for child in self.graph[node]:
                if child['node'].name != 'EndState':
                    out.add(child['node'])
        return out

    @staticmethod
    def __run_condition(info: dict, value: StrSequence):
        condition = info['edge']
        if isinstance(condition, Criteria) or isinstance(condition, Callable):
            return condition(value)
        elif isinstance(condition, list):
            out = []
            if info['args'] == all:
                quick = True
            else:
                quick = False

            for func in condition:
                if quick:
                    if not func(value):
                        return False
                else:
                    out.append(func(value))
            if quick:
                return True
            else:
                return info['args'](out)
        else:
            return True

    def run(self, value: StrSequence, verbose=False):
        # -- depth first search -- #
        vertex = self.init_state
        pre_v = vertex.name
        while vertex.name != 'EndState':
            # -- do operation -- #
            value = vertex(value)
            if verbose and vertex.name == 'Normal':
                try:
                    print(value.sentence_list[-1])
                except:
                    pass
            for child in self.graph[vertex]:
                child_node = child['node']
                if self.__run_condition(child, value):
                    vertex = child_node
                    if verbose:
                        print(value)
                        if vertex.name != pre_v:
                            print('<{}>:'.format(vertex.name))
                            pre_v = vertex.name
                    break
        value = vertex(value)
        return value



# -*- coding:utf-8 -*-
# CREATED BY: bohuai jiang 
# CREATED ON: 2020/8/17 10:01 AM
# LAST MODIFIED ON:
# AIM:

import unittest

from sentence_spliter import spliter

from sentence_spliter.logic_graph import long_short_cuter, simple_cuter
from sentence_spliter.automata.state_machine import StateMachine
from sentence_spliter.automata.sequence import StrSequence


class test_spliter(unittest.TestCase):
    def setUp(self) -> None:
        self.long_short_machine = StateMachine(long_short_cuter())
        self.simple_logic = StateMachine(simple_cuter())

    # @unittest.skip('ignore')
    def test_book(self):
        sentence = '《霸道总裁你在哪？》是一本很好看的书，小明非常喜欢这本书。《霸道总裁你在哪？》是一本很好看的书，小明非常喜欢这本书。'
        out = spliter.cut_to_sentences(sentence)
        expect = ['《霸道总裁你在哪？》是一本很好看的书，小明非常喜欢这本书。', '《霸道总裁你在哪？》是一本很好看的书，小明非常喜欢这本书。']
        self.assertEqual(expect, out)

    # @unittest.skip('ignore')
    def test_simple_quota(self):
        sentence = '"《霸道总裁你在哪？》是一本很好看的书!"，小明自豪的说到。'
        out = spliter.cut_to_sentences(sentence)
        expect = ['"《霸道总裁你在哪？》是一本很好看的书!"，小明自豪的说到。']
        self.assertEqual(expect, out)

    # @unittest.skip('ignore')
    def test_simple_bracket(self):
        sentence = '"《霸道总裁你在哪？》是一本很好看的书!"，小明自豪的说到(其实小明并不知道写的是什么。)。' \
                   '"《霸道总裁你在哪？》是一本很好看的书!"，小明自豪的说到(其实小明并不知道写的是什么。)。'
        out = spliter.cut_to_sentences(sentence)
        expect = ['"《霸道总裁你在哪？》是一本很好看的书!"，小明自豪的说到(其实小明并不知道写的是什么。)。',
                  '"《霸道总裁你在哪？》是一本很好看的书!"，小明自豪的说到(其实小明并不知道写的是什么。)。']
        self.assertEqual(expect, out)

    # @unittest.skip('ignore')
    def test_end_sentence(self):
        sentence = '在很久很久以前......... 有座山。山里有座庙啊!!!庙里竟然有个老和尚！？。。。。'
        out = self.simple_logic.run(StrSequence(sentence))
        out = out.sentence_list()
        expect = ['在很久很久以前.........', ' 有座山。', '山里有座庙啊!!!', '庙里竟然有个老和尚！？。。。。']
        self.assertEqual(expect, out)

        sentence = '在很久很久以前......... 有座山。山里有座庙啊!!!庙里竟然有个老和尚！？。。。。www.baidu.com。1.23.4'
        out = self.simple_logic.run(StrSequence(sentence))
        out = out.sentence_list()
        expect = ['在很久很久以前.........', ' 有座山。', '山里有座庙啊!!!', '庙里竟然有个老和尚！？。。。。', 'www.baidu.com。', '1.23.4']
        self.assertEqual(expect, out)

    # @unittest.skip('ignore')
    def test_quota(self):
        sentence = '张晓风笑着说道，“我们这些年可比过去强多了！“过去吃不起饭，穿不暖衣服。 现在呢？要啥有啥！'
        out = spliter.cut_to_sentences(sentence)
        expect = ['张晓风笑着说道，“我们这些年可比过去强多了！', '“过去吃不起饭，穿不暖衣服。 现在呢？要啥有啥！']
        self.assertEqual(expect, out)

    # @unittest.skip('ignore')
    def test_Lquota(self):
        sentence = '“我和你讨论的不是一个东西，死亡率与死亡比例是不同的”“你知道么？CNN你们总是制造假新闻。。。”'
        out = spliter.cut_to_sentences(sentence)
        expect = ['“我和你讨论的不是一个东西，死亡率与死亡比例是不同的”', '“你知道么？CNN你们总是制造假新闻。。。”']
        self.assertEqual(expect, out)

        sentence = '“我和你讨论的不是一个东西，死亡率与死亡比例是不同的” “你知道么？CNN你们总是制造假新闻。。。”'
        out = spliter.cut_to_sentences(sentence)
        expect = ['“我和你讨论的不是一个东西，死亡率与死亡比例是不同的” ', '“你知道么？CNN你们总是制造假新闻。。。”']
        self.assertEqual(expect, out)

    # @unittest.skip('ignore')
    def test_quota(self):
        sentence = '"师傅，前面有烟火，我去去就来！"，悟空道。"徒儿慢点，记住千万不要伤到人家"。唐生急忙跑过来。'
        out = spliter.cut_to_sentences(sentence)
        expect = ['"师傅，前面有烟火，我去去就来！"，悟空道。', '"徒儿慢点，记住千万不要伤到人家"。', '唐生急忙跑过来。']
        self.assertEqual(expect, out)

    # @unittest.skip('ignore')
    def test_long(self):
        sentence = '1234567890123456789012345678901234567890123456789012345678901234567890' \
                   '1234567890123456789012345678901234567890123456789012345678901234567890' \
                   '1234567890123456789012345678901234567890123456789012345678901234567890'
        long_short_machine = StateMachine(long_short_cuter(hard_max=70, max_len=70))
        out = long_short_machine.run(StrSequence(sentence))
        out = out.sentence_list()
        expect = ['1234567890123456789012345678901234567890123456789012345678901234567890',
                  '1234567890123456789012345678901234567890123456789012345678901234567890',
                  '1234567890123456789012345678901234567890123456789012345678901234567890']
        self.assertEqual(expect, out)

    # @unittest.skip('ignore')
    def test_short(self):
        sentence = '1234567890.1234567890。'
        out = spliter.cut_to_sentences(sentence)
        expect = ['1234567890.1234567890。']
        self.assertEqual(expect, out)

        sentence = '123。12354657789。 12324354364。'
        out = spliter.cut_to_sentences(sentence)
        expect = ['123。','12354657789。', ' 12324354364。']
        self.assertEqual(expect, out)

    def test_says(self):
        sentence = '“没走，在街上找了个客栈，闹出一场自尽的把戏，现在又上门来了。”她说'
        out = spliter.cut_to_sentences(sentence)
        expect = ['“没走，在街上找了个客栈，闹出一场自尽的把戏，现在又上门来了。”她说']
        self.assertEqual(expect, out)

    def test_empty(self):
        sentence = ''
        out = spliter.cut_to_sentences(sentence)
        expect = []
        self.assertEqual(expect, out)
    def test_long(self):
        de

if __name__ == '__main__':
    unittest.main()
